# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
import argparse
import os
import random
import time

import numpy as np
import paddle
import paddle.nn.functional as F

import paddlenlp as ppnlp
from paddlenlp.data import Stack, Tuple, Pad
from paddlenlp.datasets import load_dataset
from paddlenlp.transformers import LinearDecayWithWarmup

from visualdl import LogWriter

parser = argparse.ArgumentParser()
parser.add_argument("--save_dir", default='./checkpoint', type=str, help="The output directory where the model checkpoints will be written.")
parser.add_argument("--max_seq_length", default=16, type=int, help="The maximum total input sequence length after tokenization. "
    "Sequences longer than this will be truncated, sequences shorter will be padded.")
parser.add_argument("--batch_size", default=32, type=int, help="Batch size per GPU/CPU for training.")
parser.add_argument("--seed", type=int, default=1000, help="random seed for initialization")
parser.add_argument('--device', choices=['cpu', 'gpu', 'xpu'], default="gpu", help="Select which device to train model, defaults to gpu.")
parser.add_argument("--eval_ds", type=str, default="eval_list.txt")
parser.add_argument("--model_names", nargs="+")
parser.add_argument("--model_paths", nargs="+")
args = parser.parse_args()


def set_seed(seed):
    """sets random seed"""
    random.seed(seed)
    np.random.seed(seed)
    paddle.seed(seed)

@paddle.no_grad()
def evaluate(model, criterion, metric, data_loader):
    """
    Given a dataset, it evals model and computes the metric.

    Args:
        model(obj:`paddle.nn.Layer`): A model to classify texts.
        data_loader(obj:`paddle.io.DataLoader`): The dataset loader which generates batches.
        criterion(obj:`paddle.nn.Layer`): It can compute the loss.
        metric(obj:`paddle.metric.Metric`): The evaluation metric.
    """

    model.eval()
    metric.reset()
    losses = []
    for batch in data_loader:
        input_ids, token_type_ids, labels = batch
        logits = model(input_ids, token_type_ids)
        loss = criterion(logits, labels)
        losses.append(loss.numpy())
        correct = metric.compute(logits, labels)
        metric.update(correct)
        accu = metric.accumulate()

    print("eval loss: %.5f, accu: %.5f" % (np.mean(losses), accu))
    model.train()
    metric.reset()
    return np.mean(losses), accu


def convert_example(example, tokenizer, max_seq_length=512, is_test=False):
    """
    Builds model inputs from a sequence or a pair of sequence for sequence classification tasks
    by concatenating and adding special tokens. And creates a mask from the two sequences passed 
    to be used in a sequence-pair classification task.
        
    A BERT sequence has the following format:

    - single sequence: ``[CLS] X [SEP]``
    - pair of sequences: ``[CLS] A [SEP] B [SEP]``

    A BERT sequence pair mask has the following format:
    ::
        0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1
        | first sequence    | second sequence |

    If only one sequence, only returns the first portion of the mask (0's).


    Args:
        example(obj:`list[str]`): List of input data, containing text and label if it have label.
        tokenizer(obj:`PretrainedTokenizer`): This tokenizer inherits from :class:`~paddlenlp.transformers.PretrainedTokenizer` 
            which contains most of the methods. Users should refer to the superclass for more information regarding methods.
        max_seq_len(obj:`int`): The maximum total input sequence length after tokenization. 
            Sequences longer than this will be truncated, sequences shorter will be padded.
        is_test(obj:`False`, defaults to `False`): Whether the example contains label or not.

    Returns:
        input_ids(obj:`list[int]`): The list of token ids.
        token_type_ids(obj: `list[int]`): List of sequence pair mask.
        label(obj:`numpy.array`, data type of int64, optional): The input label if not is_test.
    """

    encoded_inputs = tokenizer(text=example["text"], max_seq_len=max_seq_length)
    input_ids = encoded_inputs["input_ids"]
    token_type_ids = encoded_inputs["token_type_ids"]

    if not is_test:
        label = np.array([example["label"]], dtype="int64")
        return input_ids, token_type_ids, label
    else:
        return input_ids, token_type_ids


def create_dataloader(dataset,
                      mode='train',
                      batch_size=1,
                      batchify_fn=None,
                      trans_fn=None):
    if trans_fn:
        dataset = dataset.map(trans_fn)

    shuffle = True if mode == 'train' else False
    if mode == 'train':
        batch_sampler = paddle.io.DistributedBatchSampler(
            dataset, batch_size=batch_size, shuffle=shuffle)
    else:
        batch_sampler = paddle.io.BatchSampler(
            dataset, batch_size=batch_size, shuffle=shuffle)

    return paddle.io.DataLoader(
        dataset=dataset,
        batch_sampler=batch_sampler,
        collate_fn=batchify_fn,
        return_list=True)

def read(data_path):
    with open(data_path, 'r', encoding='utf-8') as f:
        # 跳过列名
        next(f)
        for line in f:
            words, labels = line.strip('\n').split('\t')
            words = words.split('\002')
            labels = labels.split('\002')
            yield {'text': words[0], 'label': labels[0]}

def do_eval():
    paddle.set_device(args.device)
    rank = paddle.distributed.get_rank()
    if paddle.distributed.get_world_size() > 1:
        paddle.distributed.init_parallel_env()

    set_seed(args.seed)

    eval_res = {}
    for model_name in args.model_names:
        test_ds = load_dataset(read,data_path=args.eval_ds,lazy=False)

        if (model_name == "rbt3"):
            model = ppnlp.transformers.RobertaForSequenceClassification.from_pretrained('rbt3', num_class=2)
            tokenizer = ppnlp.transformers.RobertaTokenizer.from_pretrained('rbt3')
        elif (model_name == "bert_base_chinese"):
            model = ppnlp.transformers.BertForSequenceClassification.from_pretrained('bert-base-chinese', num_class=2)
            tokenizer = ppnlp.transformers.BertTokenizer.from_pretrained('bert-base-chinese')
        elif (model_name == "chinese_electra_small"):
            model = ppnlp.transformers.ElectraForSequenceClassification.from_pretrained('chinese-electra-small', num_classes=2)
            tokenizer = ppnlp.transformers.ElectraTokenizer.from_pretrained('chinese-electra-small')
        elif (model_name == "ernie_tiny"):
            model = ppnlp.transformers.ErnieForSequenceClassification.from_pretrained('ernie-tiny', num_classes=2)
            tokenizer = ppnlp.transformers.ErnieTinyTokenizer.from_pretrained('ernie-tiny')

        # 通过环境变量获取Artifact的存储路径
        # 具体讲解可参考：https://cloud.baidu-int.com/icloud/PaddleCloud/%E6%93%8D%E4%BD%9C%E6%8C%87%E5%8D%97/pipeline%E6%93%8D%E4%BD%9C%E6%8C%87%E5%8D%97/%E7%BC%96%E6%8E%92/%E8%AF%A6%E8%A7%A3#%E9%80%9A%E8%BF%87%E7%8E%AF%E5%A2%83%E5%8F%98%E9%87%8F%E8%8E%B7%E5%8F%96
        model_dir = os.getenv(model_name)

        # model_dir是一个表示路径的字符串，如"./rbt3"
        state_dict = paddle.load(model_dir + "/model_state.pdparams")
        model.set_dict(state_dict)
        model = paddle.DataParallel(model)

        trans_func = partial(
            convert_example,
            tokenizer=tokenizer,
            max_seq_length=args.max_seq_length)

        batchify_fn = lambda samples, fn=Tuple(
            Pad(axis=0, pad_val=tokenizer.pad_token_id),  # input
            Pad(axis=0, pad_val=tokenizer.pad_token_type_id),  # segment
            Stack(dtype="int64")  # label
        ): [data for data in fn(samples)]

        test_data_loader = create_dataloader(
            test_ds,
            mode='test',
            batch_size=args.batch_size,
            batchify_fn=batchify_fn,
            trans_fn=trans_func)

        criterion = paddle.nn.loss.CrossEntropyLoss()
        metric = paddle.metric.Accuracy()
        
        # 保存评估结果
        print("test result of " + model_name)
        eval_loss, eval_acc = evaluate(model, criterion, metric, test_data_loader)
        eval_res[model_name] = (model, tokenizer, eval_loss, eval_acc)

    best_accu = 0
    best_model_name = ""
    for name in eval_res.keys():
        if best_accu < eval_res[name][3]:
            best_accu = eval_res[name][3]
            best_model_name = name

    # 保存模型
    save_dir = args.save_dir
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
    eval_res[best_model_name][0]._layers.save_pretrained(save_dir)
    eval_res[best_model_name][1].save_pretrained(save_dir)
    with open(save_dir+"/model_name", "w") as f:
        f.write(best_model_name)


if __name__ == "__main__":
    do_eval()